@@ -5,21 +5,20 @@
 from __future__ import annotations
 
 import os
-from pathlib import Path
-from warnings import warn
-from typing import Callable, Optional, Union, List
-
-try:
-    from functools import cached_property
-except ImportError:
-    from cached_property import cached_property  # type: ignore
-
-import spacy
-from spacy.tokens.doc import Doc
-import pandas as pd
+import sqlite3
 from glob import glob
-from toolz import compose, partial, identity
+from pathlib import Path
+from typing import Any, Callable, Optional, Sequence, Union
+from warnings import warn
 
+import numpy as np
+import pandas as pd
+import spacy
+from spacy.tokens import Token
+from spacy.tokens.doc import Doc
+from toolz import compose, identity, memoize, partial
+
+import textnets as tn
 
 #: Mapping of language codes to spacy language model names.
 LANGS = {
@@ -29,6 +28,7 @@ LANGS = {
     "el": "el_core_news_sm",  # Greek
     "en": "en_core_web_sm",  # English
     "es": "es_core_news_sm",  # Spanish
+    "fi": "fi_core_news_sm",  # Finnish
     "fr": "fr_core_news_sm",  # French
     "it": "it_core_news_sm",  # Italian
     "ja": "ja_core_news_sm",  # Japanese
@@ -40,9 +40,15 @@ LANGS = {
     "pt": "pt_core_news_sm",  # Portuguese
     "ro": "ro_core_news_sm",  # Romanian
     "ru": "ru_core_news_sm",  # Russian
+    "sv": "sv_core_news_sm",  # Swedish
     "zh": "zh_core_web_sm",  # Chinese
 }
 
+#: Custom type for objects resembling documents (token sequences).
+DocLike = Union[Doc, Sequence[Token]]
+
+_INSTALLED_MODELS = spacy.util.get_installed_models()
+
 
 class Corpus:
     """
@@ -54,36 +60,61 @@ class Corpus:
         Series containing the documents. The index must contain document
         labels.
     lang : str, optional
-        The langugage model to use (default: ``en_core_web_sm``).
+        The langugage model to use (default set by "lang" parameter).
+
+    Raises
+    ------
+    ValueError
+        If the supplied data is empty.
+
+    Attributes
+    ----------
+    documents : Series
+        The corpus documents.
+    lang : str
+        The language model used (ISO code or spaCy model name).
     """
 
     def __init__(
         self,
         data: pd.Series,
-        lang: str = "en_core_web_sm",
+        lang: Optional[str] = None,
     ) -> None:
         if data.empty:
-            raise ValueError("Corpus data is empty")
+            raise ValueError("Corpus data is empty.")
         documents = data.copy()
-        if documents.isna().any():
-            missings = documents.isna().sum()
+        if missings := documents.isna().sum():
             warn(f"Dropping {missings} empty document(s).")
             documents = documents[~documents.isna()]
         documents.index = documents.index.set_names(["label"])
         self.documents = documents
-        self.lang = LANGS[lang] if lang in LANGS.keys() else lang
+        if lang is None:
+            lang = tn.params["lang"]
+        self.lang = LANGS.get(lang, lang)
+        if self.lang not in _INSTALLED_MODELS:
+            warn(f"Language model '{self.lang}' is not yet installed.")
 
-    @cached_property
+    @property
     def nlp(self) -> pd.Series:
         """Corpus documents with NLP applied."""
+        return self._nlp(self.lang)
+
+    @memoize
+    def _nlp(self, lang: str) -> pd.Series:
         try:
-            nlp = spacy.load(self.lang, disable=["ner", "textcat"])
+            nlp = spacy.load(lang, disable=["ner", "textcat"])
         except OSError as err:
-            if self.lang in LANGS.values():
+            if tn.params["autodownload"]:
+                try:
+                    spacy.cli.download(lang)  # type: ignore
+                    _INSTALLED_MODELS.append(lang)
+                    return self._nlp(lang)
+                except (KeyError, OSError):
+                    pass
+            elif lang in LANGS.values():
                 raise err
-            else:
-                nlp = spacy.blank(self.lang)
-                warn(f"Using basic {self.lang} language model.")
+            nlp = spacy.blank(lang)
+            warn(f"Using basic '{lang}' language model.")
         return self.documents.map(_normalize_whitespace).map(nlp)
 
     def __len__(self) -> int:
@@ -97,7 +128,7 @@ class Corpus:
         cls,
         data: pd.DataFrame,
         doc_col: Optional[str] = None,
-        lang: str = "en_core_web_sm",
+        lang: Optional[str] = None,
     ) -> Corpus:
         """
         Create corpus from data frame.
@@ -105,32 +136,59 @@ class Corpus:
         Parameters
         ----------
         data : DataFrame
-            Series containing the documents. The index must contain document
+            DataFrame containing documents. The index must contain document
             labels.
         doc_col : str, optional
-            If ``data`` is a data frame, this indicates which column contains the
-            document texts. If none is specified, the first column with strings is
-            used.
+            Indicates which column of ``data`` contains the document texts. If
+            none is specified, the first column with strings is used.
         lang : str, optional
-            The langugage model to use (default: ``en_core_web_sm``).
+            The langugage model to use (default set by "lang" parameter).
+
+        Raises
+        ------
+        NoDocumentColumnException
+            If no document column can be detected.
 
         Returns
         -------
-        Corpus
+        `Corpus`
         """
         object_cols = data.select_dtypes(include="object").columns
         if not doc_col and object_cols.empty:
             raise NoDocumentColumnException("No suitable document column.")
-        elif not doc_col:
+        if not doc_col:
             doc_col = object_cols[0]
         return cls(data.copy()[doc_col], lang=lang)
 
+    @classmethod
+    def from_dict(
+        cls,
+        data: dict[Any, str],
+        lang: Optional[str] = None,
+    ) -> Corpus:
+        """
+        Create corpus from dictionary.
+
+        Parameters
+        ----------
+        data : dict
+            Dictionary containing the documents as values and document labels
+            as keys.
+        lang : str, optional
+            The langugage model to use (default set by "lang" parameter).
+
+        Returns
+        -------
+        `Corpus`
+        """
+        return cls(pd.Series(data), lang=lang)
+
     @classmethod
     def from_files(
         cls,
-        files: Union[str, List[str], List[Path]],
-        doc_labels: Optional[List[str]] = None,
-        lang: str = "en_core_web_sm",
+        files: Union[str, list[str], list[Path]],
+        doc_labels: Optional[list[str]] = None,
+        lang: Optional[str] = None,
     ) -> Corpus:
         """Construct corpus from files.
 
@@ -141,24 +199,31 @@ class Corpus:
         doc_labels : list of str, optional
             Labels for documents (default: file name without suffix).
         lang : str, optional
-            The langugage model to use (default: ``en_core_web_sm``).
+            The langugage model to use (default set by "lang" parameter).
+
+        Raises
+        ------
+        IsADirectoryError
+            If the provided path is a directory. (Use globbing.)
+        FileNotFoundError
+            If the provided path does not exist.
 
         Returns
         -------
-        Corpus
+        `Corpus`
         """
         if isinstance(files, str):
             files = glob(os.path.expanduser(files))
         files = [Path(f) for f in files]
-        for f in files:
-            if f.expanduser().is_file():
+        for file in files:
+            if file.expanduser().is_file():
                 pass
-            elif f.expanduser().exists():
-                raise IsADirectoryError(f.name)
+            elif file.expanduser().exists():
+                raise IsADirectoryError(file.name)
             else:
-                raise FileNotFoundError(f.name)
+                raise FileNotFoundError(file.name)
         if not doc_labels:
-            doc_labels = [f.stem for f in files]
+            doc_labels = [file.stem for file in files]
         data = pd.DataFrame({"path": files}, index=doc_labels)
         data["raw"] = data["path"].map(_read_file)
         return cls.from_df(data, doc_col="raw", lang=lang)
@@ -169,7 +234,7 @@ class Corpus:
         path: str,
         label_col: Optional[str] = None,
         doc_col: Optional[str] = None,
-        lang: str = "en_core_web_sm",
+        lang: Optional[str] = None,
         **kwargs,
     ) -> Corpus:
         """Read corpus from comma-separated value file.
@@ -185,13 +250,13 @@ class Corpus:
             Column that contains document text (default: None, in which case
             the first text column is used).
         lang : str, optional
-            The langugage model to use (default: ``en_core_web_sm``).
+            The langugage model to use (default set by "lang" parameter).
         kwargs
             Arguments to pass to `pandas.read_csv`.
 
         Returns
         -------
-        Corpus
+        `Corpus`
         """
         kwargs.setdefault("index_col", label_col)
         data = pd.read_csv(path, **kwargs)
@@ -206,7 +271,7 @@ class Corpus:
         conn: Union[str, object],
         label_col: Optional[str] = None,
         doc_col: Optional[str] = None,
-        lang: str = "en_core_web_sm",
+        lang: Optional[str] = None,
         **kwargs,
     ) -> Corpus:
         """Read corpus from SQL database.
@@ -224,13 +289,13 @@ class Corpus:
             Column that contains document text (default: None, in which case
             the first text column is used).
         lang : str, optional
-            The langugage model to use (default: ``en_core_web_sm``).
+            The langugage model to use (default set by "lang" parameter).
         kwargs
             Arguments to pass to `pandas.read_sql`.
 
         Returns
         -------
-        Corpus
+        `Corpus`
         """
         kwargs.setdefault("index_col", label_col)
         data = pd.read_sql(qry, conn, **kwargs)
@@ -238,16 +303,67 @@ class Corpus:
             data = data.set_index(data.columns[0])
         return cls.from_df(data, doc_col=doc_col, lang=lang)
 
+    def save(self, target: Union[os.PathLike[Any], str]) -> None:
+        """
+        Save a corpus to file.
+
+        Parameters
+        ----------
+        target : str or path
+            File to save the corpus to. If the file exists, it will be
+            overwritten.
+        """
+        conn = sqlite3.connect(Path(target))
+        meta = {"lang": self.lang}
+        with conn as c:
+            self.documents.to_sql("corpus_documents", c, if_exists="replace")
+            pd.Series(meta, name="values").to_sql(
+                "corpus_meta", c, if_exists="replace", index_label="keys"
+            )
+
+    @classmethod
+    def load(cls, source: Union[os.PathLike[Any], str]) -> Corpus:
+        """
+        Load a corpus from file.
+
+        Parameters
+        ----------
+        source : str or path
+            File to read the corpus from. This should be a file created by
+            `Corpus.save`.
+
+        Raises
+        ------
+        FileNotFoundError
+            If the specified path does not exist.
+
+        Returns
+        -------
+        `Corpus`
+        """
+        if not Path(source).exists():
+            raise FileNotFoundError(f"File '{source}' does not exist.")
+        conn = sqlite3.connect(Path(source))
+        with conn:
+            documents = pd.read_sql(
+                "SELECT * FROM corpus_documents", conn, index_col="label"
+            )
+            meta = pd.read_sql("SELECT * FROM corpus_meta", conn, index_col="keys")[
+                "values"
+            ]
+        return cls.from_df(documents, lang=meta["lang"])
+
     def tokenized(
         self,
-        remove: List[str] = [],
+        remove: Optional[list[str]] = None,
         stem: bool = True,
         remove_stop_words: bool = True,
         remove_urls: bool = True,
         remove_numbers: bool = True,
         remove_punctuation: bool = True,
         lower: bool = True,
-    ) -> pd.DataFrame:
+        sublinear: bool = True,
+    ) -> TidyText:
         """Return tokenized version of corpus in tidy format.
 
         Parameters
@@ -267,15 +383,20 @@ class Corpus:
             (default: True).
         lower : bool, optional
             Make lower-case (default: True).
+        sublinear : bool, optional
+            Apply sublinear scaling when calculating *tf-idf* term weights
+            (default: True).
 
         Returns
         -------
-        pd.DataFrame
+        `pandas.DataFrame`
             A data frame with document labels (index), tokens (term), and
             per-document counts (n).
         """
         func = compose(
-            partial(_remove_additional, token_list=remove) if remove else identity,
+            partial(_remove_additional, token_list=remove)
+            if remove is not None
+            else identity,
             _lower if lower else identity,
             _stem if stem else _as_text,
             _remove_stop_words if remove_stop_words else identity,
@@ -283,11 +404,15 @@ class Corpus:
             _remove_numbers if remove_numbers else identity,
             _remove_punctuation if remove_punctuation else identity,
         )
-        return self._return_tidy_text(func)
+        tt = self._make_tidy_text(func)
+        return _tf_idf(tt, sublinear)
 
     def noun_phrases(
-        self, normalize: bool = False, remove: List[str] = []
-    ) -> pd.DataFrame:
+        self,
+        normalize: bool = False,
+        remove: Optional[list[str]] = None,
+        sublinear: bool = True,
+    ) -> TidyText:
         """Return noun phrases from corpus in tidy format.
 
         Parameters
@@ -296,30 +421,37 @@ class Corpus:
             Return lemmas of noun phrases (default: False).
         remove : list of str, optional
             Additional tokens to remove.
+        sublinear : bool, optional
+            Apply sublinear scaling when calculating *tf-idf* term weights
+            (default: True).
 
         Returns
         -------
-        pd.DataFrame
+        `pandas.DataFrame`
             A data frame with document labels (index), noun phrases
             (term), and per-document counts (n).
         """
         func = compose(
-            partial(_remove_additional, token_list=remove) if remove else identity,
+            partial(_remove_additional, token_list=remove)
+            if remove is not None
+            else identity,
             partial(_noun_chunks, normalize=normalize),
         )
-        return self._return_tidy_text(func)
+        tt = self._make_tidy_text(func)
+        return _tf_idf(tt, sublinear)
 
     def ngrams(
         self,
         size: int,
-        remove: List[str] = [],
+        remove: Optional[list[str]] = None,
         stem: bool = False,
         remove_stop_words: bool = False,
         remove_urls: bool = False,
         remove_numbers: bool = False,
         remove_punctuation: bool = False,
         lower: bool = False,
-    ) -> pd.DataFrame:
+        sublinear: bool = True,
+    ) -> TidyText:
         """Return n-grams of length n from corpus in tidy format.
 
         Parameters
@@ -341,16 +473,21 @@ class Corpus:
             (default: False).
         lower : bool, optional
             Make lower-case (default: False).
+        sublinear : bool, optional
+            Apply sublinear scaling when calculating *tf-idf* term weights
+            (default: True).
 
         Returns
         -------
-        pd.DataFrame
+        `pandas.DataFrame`
             A data frame with document labels (index), n-grams (term), and
             per-document counts (n).
         """
         func = compose(
             partial(_ngrams, n=size),
-            partial(_remove_additional, token_list=remove) if remove else identity,
+            partial(_remove_additional, token_list=remove)
+            if remove is not None
+            else identity,
             _lower if lower else identity,
             _stem if stem else _as_text,
             _remove_stop_words if remove_stop_words else identity,
@@ -358,10 +495,11 @@ class Corpus:
             _remove_numbers if remove_numbers else identity,
             _remove_punctuation if remove_punctuation else identity,
         )
-        return self._return_tidy_text(func)
+        tt = self._make_tidy_text(func)
+        return _tf_idf(tt, sublinear)
 
-    def _return_tidy_text(self, func: Callable[[Doc], List[str]]) -> pd.DataFrame:
-        return (
+    def _make_tidy_text(self, func: Callable[[Doc], list[str]]) -> TidyText:
+        tt = (
             pd.melt(
                 self.nlp.map(func).apply(pd.Series).reset_index(),
                 id_vars="label",
@@ -373,6 +511,13 @@ class Corpus:
             .reset_index()
             .set_index("label")
         )
+        return TidyText(tt)
+
+    def __repr__(self) -> str:
+        return (
+            f"<{self.__class__.__name__} with {len(self.documents)} documents "
+            + f"using language model '{self.lang}'>"
+        )
 
     def _repr_html_(self) -> str:
         tbl = pd.DataFrame(self.documents).to_html(
@@ -394,7 +539,6 @@ class Corpus:
                 margin-bottom: 1em;
               }}
             </style>
-            {tbl}
             <details>
               <summary>
                 <table class="full-width">
@@ -414,7 +558,7 @@ class Corpus:
                   </tr>
                 </table>
               </summary>
-              <pre>{self.__doc__}</pre>
+              {tbl}
             </details>"""
 
 
@@ -423,12 +567,12 @@ def _read_file(file_name: Path) -> str:
     return file_name.read_bytes().decode("utf-8", "replace").strip()
 
 
-def _normalize_whitespace(s: str) -> str:
+def _normalize_whitespace(string: str) -> str:
     """Replace all whitespace with single spaces."""
-    return " ".join(s.split())
+    return " ".join(string.split())
 
 
-def _noun_chunks(doc: Doc, normalize) -> List[str]:
+def _noun_chunks(doc: Doc, normalize: bool) -> list[str]:
     """Return only the noun chunks in lower case."""
     return [
         (chunk.lemma_ if normalize else " ".join([t.lower_ for t in chunk]))
@@ -437,22 +581,22 @@ def _noun_chunks(doc: Doc, normalize) -> List[str]:
     ]
 
 
-def _remove_stop_words(doc: Doc) -> Doc:
+def _remove_stop_words(doc: DocLike) -> list[Token]:
     """Return document without stop words."""
     return [word for word in doc if not word.is_stop]
 
 
-def _remove_urls(doc: Doc) -> Doc:
+def _remove_urls(doc: DocLike) -> list[Token]:
     """Return document without URLs or email addresses."""
     return [word for word in doc if not word.like_url and not word.like_email]
 
 
-def _remove_numbers(doc: Doc) -> Doc:
+def _remove_numbers(doc: DocLike) -> list[Token]:
     """Return document without numbers."""
     return [word for word in doc if not word.like_num]
 
 
-def _remove_punctuation(doc: Doc) -> Doc:
+def _remove_punctuation(doc: DocLike) -> list[Token]:
     """Return document without punctuation, brackets and quotation marks."""
     return [
         word
@@ -461,30 +605,55 @@ def _remove_punctuation(doc: Doc) -> Doc:
     ]
 
 
-def _stem(doc: Doc) -> List[str]:
+def _stem(doc: DocLike) -> list[str]:
     """Return list of word stem strings."""
     return [word.lemma_ for word in doc]
 
 
-def _as_text(doc: Doc) -> List[str]:
+def _as_text(doc: DocLike) -> list[str]:
     """Turn document into list of strings."""
     return [word.text for word in doc]
 
 
-def _lower(doc: List[str]) -> List[str]:
+def _lower(doc: list[str]) -> list[str]:
     """Return list of strings in lower case."""
     return [s.lower() for s in doc]
 
 
-def _remove_additional(doc: List[str], token_list: List[str]) -> List[str]:
+def _remove_additional(doc: list[str], token_list: list[str]) -> list[str]:
     """Return list of strings without specified tokens."""
     return [s for s in doc if s not in token_list]
 
 
-def _ngrams(doc: List[str], n: int) -> List[str]:
-    """Returns list of n-gram strings."""
+def _ngrams(doc: list[str], n: int) -> list[str]:
+    """Return list of n-gram strings."""
     return [" ".join(t) for t in zip(*[doc[offset:] for offset in range(n)])]
 
 
+def _tf_idf(tidy_text: TidyText, sublinear: bool) -> TidyText:
+    """Calculate term frequency/inverse document frequency."""
+    if sublinear:
+        tidy_text["tf"] = tidy_text["n"].map(_sublinear_scaling)
+    else:
+        totals = tidy_text.groupby(tidy_text.index).sum().rename(columns={"n": "total"})
+        tidy_text = tidy_text.merge(totals, right_index=True, left_index=True)
+        tidy_text["tf"] = tidy_text["n"] / tidy_text["total"]
+    idfs = np.log10(len(set(tidy_text.index)) / tidy_text["term"].value_counts())
+    tt = tidy_text.merge(pd.DataFrame(idfs), left_on="term", right_index=True).rename(
+        columns={"term_y": "idf"}
+    )
+    tt["term_weight"] = tt["tf"] * tt["idf"]
+    return tt[["term", "n", "term_weight"]]
+
+
+def _sublinear_scaling(n: Union[int, float]) -> float:
+    """Logarithmic scaling function."""
+    return 1 + np.log10(n) if n > 0 else 0
+
+
 class NoDocumentColumnException(Exception):
     """Raised if no suitable document column is specified or found."""
+
+
+class TidyText(pd.DataFrame):
+    """Collection of tokens with per-document counts."""
